"""CLI integration."""
import typing as t

import os
import re
import sys

import click
from playhouse.db_url import connect

from . import LOGGER, MIGRATE_TABLE
from .router import Router


CLEAN_RE: t.Pattern = re.compile(r'\s+$', re.M)
VERBOSE: t.List[str] = ['WARNING', 'INFO', 'DEBUG', 'NOTSET']


def get_router(directory: str = None, database: str = None,
               migratetable: str = MIGRATE_TABLE, verbose: int = 0) -> Router:
    """Load and initialize a router."""
    config: t.Dict = {}
    logging_level: str = VERBOSE[verbose]
    ignore = schema = None

    if directory:
        try:
            with open(os.path.join(directory, 'conf.py')) as cfg:
                code = compile(cfg.read(), '<string>', 'exec', dont_inherit=True)
                exec(code, config, config)
                database = config.get('DATABASE', database)
                ignore = config.get('IGNORE', ignore)
                schema = config.get('SCHEMA', schema)
                migratetable = config.get('MIGRATE_TABLE', migratetable)
                logging_level = config.get('LOGGING_LEVEL', logging_level).upper()
        except IOError:
            pass

    if isinstance(database, str):
        database = connect(database)

    LOGGER.setLevel(logging_level)

    if not database:
        LOGGER.error("Database is undefined")
        return sys.exit(1)

    try:
        return Router(database, migrate_table=migratetable, migrate_dir=directory,
                      ignore=ignore, schema=schema)
    except RuntimeError as exc:
        LOGGER.error(exc)
        return sys.exit(1)


@click.group()
def cli():
    """Just a group."""
    pass


@cli.command()
@click.option('--name', default=None, help="Select migration")
@click.option('--database', default=None, help="Database connection")
@click.option('--directory', default='migrations', help="Directory where migrations are stored")
@click.option('--fake', is_flag=True, default=False, help="Run migration as fake.")
@click.option('--migratetable', default="migratehistory", help="Migration table.")
@click.option('-v', '--verbose', count=True)
def migrate(name: str = None, database: str = None, directory: str = None,
            migratetable: str = MIGRATE_TABLE, verbose: int = 0, fake: bool = False):
    """Migrate database."""
    router = get_router(directory, database, migratetable, verbose)
    migrations = router.run(name, fake=fake)
    if migrations:
        click.echo('Migrations completed: %s' % ', '.join(migrations))


@cli.command()
@click.argument('name')
@click.option('--auto', default=False, is_flag=True, help=(
    "Scan sources and create db migrations automatically. "
    "Supports autodiscovery."))
@click.option('--auto-source', default=None, help=(
    "Set to python module path for changes autoscan (e.g. 'package.models'). "
    "Current directory will be recursively scanned by default."))
@click.option('--database', default=None, help="Database connection")
@click.option('--directory', default='migrations', help="Directory where migrations are stored")
@click.option('--migratetable', default="migratehistory", help="Migration table.")
@click.option('-v', '--verbose', count=True)
def create(name: str = None, database: str = None, directory: str = None,
           migratetable: str = MIGRATE_TABLE, verbose: int = 0,
           auto: bool = False, auto_source: str = None):
    """Create a migration."""
    router: Router = get_router(directory, database, migratetable, verbose)
    router.create(name or 'auto', auto=auto_source if auto and auto_source else auto)


@cli.command()
@click.argument('name', required=False)
@click.option('--count', required=False, default=1, type=int,
              help="Number of last migrations to be rolled back."
                   "Ignored in case of non-empty name")
@click.option('--database', default=None, help="Database connection")
@click.option('--directory', default='migrations', help="Directory where migrations are stored")
@click.option('--migratetable', default="migratehistory", help="Migration table.")
@click.option('-v', '--verbose', count=True)
def rollback(name: str = None, database: str = None, directory: str = None,
             migratetable: str = MIGRATE_TABLE, verbose: int = 0, count: int = 1):
    """Rollback a migration with given name or number of last migrations with given --count option as integer number"""  # noqa
    router: Router = get_router(directory, database, migratetable, verbose)
    if not name:
        if len(router.done) < count:
            raise RuntimeError(
                'Unable to rollback %s migrations from %s: %s' %
                (count, len(router.done), router.done))
        for _ in range(count):
            router = get_router(directory, database, migratetable, verbose)
            name = router.done[-1]
            router.rollback(name)
    else:
        router.rollback(name)


@cli.command()
@click.option('--database', default=None, help="Database connection")
@click.option('--directory', default='migrations', help="Directory where migrations are stored")
@click.option('--migratetable', default="migratehistory", help="Migration table.")
@click.option('-v', '--verbose', count=True)
def list(database: str = None, directory: str = None,
         migratetable: str = MIGRATE_TABLE, verbose: int = 0):
    """List migrations."""
    router: Router = get_router(directory, database, migratetable, verbose)
    click.echo('Migrations are done:')
    click.echo('\n'.join(router.done))
    click.echo('')
    click.echo('Migrations are undone:')
    click.echo('\n'.join(router.diff))


@cli.command()
@click.option('--database', default=None, help="Database connection")
@click.option('--directory', default='migrations', help="Directory where migrations are stored")
@click.option('--migratetable', default="migratehistory", help="Migration table.")
@click.option('-v', '--verbose', count=True)
def merge(database: str = None, directory: str = None,
          migratetable: str = MIGRATE_TABLE, verbose: int = 0):
    """Merge migrations into one."""
    router: Router = get_router(directory, database, migratetable, verbose)
    router.merge()



"""Run migrations."""

import typing as t
from functools import wraps

import peewee as pw
from playhouse.migrate import SQL, Context, MySQLDatabase
from playhouse.migrate import MySQLMigrator as MqM
from playhouse.migrate import Operation, PostgresqlDatabase
from playhouse.migrate import PostgresqlMigrator as PgM
from playhouse.migrate import SchemaMigrator as ScM
from playhouse.migrate import SqliteDatabase
from playhouse.migrate import SqliteMigrator as SqM
from playhouse.migrate import make_index_name, operation

from peewee_migrate import LOGGER


class SchemaMigrator(ScM):
    """Implement migrations."""

    @classmethod
    def from_database(cls, database: pw.Database) -> ScM:
        """Initialize migrator by db."""
        if isinstance(database, PostgresqlDatabase):
            return PostgresqlMigrator(database)

        if isinstance(database, SqliteDatabase):
            return SqliteMigrator(database)

        if isinstance(database, MySQLDatabase):
            return MySQLMigrator(database)

        return super(SchemaMigrator, cls).from_database(database)

    def drop_table(self, model: pw.Model, cascade: bool = True) -> t.Callable:
        """Drop table."""
        return lambda: model.drop_table(cascade=cascade)

    @operation
    def change_column(
        self, table: str, column_name: str, field: pw.Field
    ) -> t.List[Operation]:
        """Change column."""
        operations = [self.alter_change_column(table, column_name, field)]
        if not field.null:
            operations.extend([self.add_not_null(table, column_name)])
        return operations

    def alter_change_column(self, table: str, column: str, field: pw.Field) -> Context:
        """Support change columns."""
        ctx = self.make_context()
        field_null, field.null = field.null, True
        ctx = (
            self._alter_table(ctx, table).literal(" ALTER COLUMN ").sql(field.ddl(ctx))
        )
        field.null = field_null
        return ctx

    @operation
    def sql(self, sql: str, *params) -> SQL:
        """Execute raw SQL."""
        return SQL(sql, *params)

    def alter_add_column(
        self, table: str, column_name: str, field: pw.Field, **kwargs
    ) -> Operation:
        """Fix fieldname for ForeignKeys."""
        name = field.name
        op = super(SchemaMigrator, self).alter_add_column(
            table, column_name, field, **kwargs
        )
        if isinstance(field, pw.ForeignKeyField):
            field.name = name
        return op


class MySQLMigrator(SchemaMigrator, MqM):
    """Support MySQL."""

    def alter_change_column(self, table: str, column: str, field: pw.Field) -> Context:
        """Support change columns."""
        ctx = self.make_context()
        field_null, field.null = field.null, True
        ctx = (
            self._alter_table(ctx, table).literal(" MODIFY COLUMN ").sql(field.ddl(ctx))
        )
        field.null = field_null
        return ctx


class PostgresqlMigrator(SchemaMigrator, PgM):
    """Support the migrations in postgresql."""

    def alter_change_column(
        self, table: str, column_name: str, field: pw.Field
    ) -> Context:
        """Support change columns."""
        context = super(PostgresqlMigrator, self).alter_change_column(
            table, column_name, field
        )
        context._sql.insert(-1, "TYPE")
        context._sql.insert(-1, " ")
        return context


class SqliteMigrator(SchemaMigrator, SqM):
    """Support the migrations in sqlite."""

    def drop_table(self, model: pw.Model, cascade: bool = True) -> t.Callable:
        """Sqlite doesnt support cascade syntax by default."""
        return lambda: model.drop_table(cascade=False)

    def alter_change_column(
        self, table: str, column: str, field: pw.Field
    ) -> Operation:
        """Support change columns."""
        return self._update_column(table, column, lambda a, b: b)


def get_model(method):
    """Convert string to model class."""

    @wraps(method)
    def wrapper(migrator, model, *args, **kwargs):
        if isinstance(model, str):
            return method(migrator, migrator.orm[model], *args, **kwargs)
        return method(migrator, model, *args, **kwargs)

    return wrapper


class Migrator(object):
    """Provide migrations."""

    def __init__(self, database: t.Union[pw.Database, pw.Proxy]):
        """Initialize the migrator."""
        if isinstance(database, pw.Proxy):
            database = database.obj

        self.database = database
        self.orm: t.Dict[str, pw.Model] = {}
        self.ops: t.List[Operation] = []
        self.migrator = SchemaMigrator.from_database(self.database)

    def run(self):
        """Run operations."""
        for op in self.ops:
            if isinstance(op, Operation):
                LOGGER.info("%s %s", op.method, op.args)
                op.run()
            else:
                op()
        self.clean()

    def python(self, func: t.Callable, *args, **kwargs):
        """Run python code."""
        self.ops.append(lambda: func(*args, **kwargs))

    def sql(self, sql: str, *params):
        """Execure raw SQL."""
        self.ops.append(self.migrator.sql(sql, *params))

    def clean(self):
        """Clean the operations."""
        self.ops = list()

    def create_table(self, model: pw.Model) -> pw.Model:
        """Create model and table in database.

        >> migrator.create_table(model)
        """
        self.orm[model._meta.table_name] = model
        model._meta.database = self.database
        self.ops.append(model.create_table)
        return model

    create_model = create_table

    @get_model
    def drop_table(self, model: pw.Model, cascade: bool = True):
        """Drop model and table from database.

        >> migrator.drop_table(model, cascade=True)
        """
        del self.orm[model._meta.table_name]
        self.ops.append(self.migrator.drop_table(model, cascade))

    remove_model = drop_table

    @get_model
    def add_columns(self, model: pw.Model, **fields: pw.Field) -> pw.Model:
        """Create new fields."""
        for name, field in fields.items():
            model._meta.add_field(name, field)
            self.ops.append(
                self.migrator.add_column(
                    model._meta.table_name, field.column_name, field
                )
            )
            if field.unique:
                self.ops.append(
                    self.migrator.add_index(
                        model._meta.table_name, (field.column_name,), unique=True
                    )
                )
        return model

    add_fields = add_columns

    @get_model
    def change_columns(self, model: pw.Model, **fields: pw.Field) -> pw.Model:
        """Change fields."""
        for name, field in fields.items():
            old_field = model._meta.fields.get(name, field)
            old_column_name = old_field and old_field.column_name

            model._meta.add_field(name, field)

            if isinstance(old_field, pw.ForeignKeyField):
                self.ops.append(
                    self.migrator.drop_foreign_key_constraint(
                        model._meta.table_name, old_column_name
                    )
                )

            if old_column_name != field.column_name:
                self.ops.append(
                    self.migrator.rename_column(
                        model._meta.table_name, old_column_name, field.column_name
                    )
                )

            if isinstance(field, pw.ForeignKeyField):
                on_delete = field.on_delete if field.on_delete else "RESTRICT"
                on_update = field.on_update if field.on_update else "RESTRICT"
                self.ops.append(
                    self.migrator.add_foreign_key_constraint(
                        model._meta.table_name,
                        field.column_name,
                        field.rel_model._meta.table_name,
                        field.rel_field.name,
                        on_delete,
                        on_update,
                    )
                )
                continue

            self.ops.append(
                self.migrator.change_column(
                    model._meta.table_name, field.column_name, field
                )
            )

            if field.unique == old_field.unique:
                continue

            if field.unique:
                index = (field.column_name,), field.unique
                self.ops.append(self.migrator.add_index(model._meta.table_name, *index))
                model._meta.indexes.append(index)
            else:
                index = (field.column_name,), old_field.unique
                self.ops.append(
                    self.migrator.drop_index(model._meta.table_name, *index)
                )
                model._meta.indexes.remove(index)

        return model

    change_fields = change_columns

    @get_model
    def drop_columns(self, model: pw.Model, *names: str, **kwargs) -> pw.Model:
        """Remove fields from model."""
        fields = [field for field in model._meta.fields.values() if field.name in names]
        cascade = kwargs.pop("cascade", True)
        for field in fields:
            self.__del_field__(model, field)
            if field.unique:
                index_name = make_index_name(
                    model._meta.table_name, [field.column_name]
                )
                self.ops.append(
                    self.migrator.drop_index(model._meta.table_name, index_name)
                )
            self.ops.append(
                self.migrator.drop_column(
                    model._meta.table_name, field.column_name, cascade=cascade
                )
            )
        return model

    remove_fields = drop_columns

    def __del_field__(self, model: pw.Model, field: pw.Field):
        """Delete field from model."""
        model._meta.remove_field(field.name)
        delattr(model, field.name)
        if isinstance(field, pw.ForeignKeyField):
            obj_id_name = field.column_name
            if field.column_name == field.name:
                obj_id_name += "_id"
            if hasattr(model, obj_id_name):
                delattr(model, obj_id_name)
            delattr(field.rel_model, field.backref)

    @get_model
    def rename_column(self, model: pw.Model, old_name: str, new_name: str) -> pw.Model:
        """Rename field in model."""
        field = model._meta.fields[old_name]
        if isinstance(field, pw.ForeignKeyField):
            old_name = field.column_name
        self.__del_field__(model, field)
        field.name = field.column_name = new_name
        if isinstance(field, pw.ForeignKeyField):
            field.column_name = field.column_name + "_id"
        model._meta.add_field(new_name, field)
        self.ops.append(
            self.migrator.rename_column(
                model._meta.table_name, old_name, field.column_name
            )
        )
        return model

    rename_field = rename_column

    @get_model
    def rename_table(self, model: pw.Model, new_name: str) -> pw.Model:
        """Rename table in database."""
        old_name = model._meta.table_name
        del self.orm[model._meta.table_name]
        model._meta.table_name = new_name
        self.orm[model._meta.table_name] = model
        self.ops.append(self.migrator.rename_table(old_name, new_name))
        return model

    @get_model
    def add_index(self, model: pw.Model, *columns: str, **kwargs) -> pw.Model:
        """Create indexes."""
        unique = kwargs.pop("unique", False)
        model._meta.indexes.append((columns, unique))
        columns_ = []
        for col in columns:
            field = model._meta.fields.get(col)

            if len(columns) == 1:
                field.unique = unique
                field.index = not unique

            if isinstance(field, pw.ForeignKeyField):
                col = col + "_id"

            columns_.append(col)
        self.ops.append(
            self.migrator.add_index(model._meta.table_name, columns_, unique=unique)
        )
        return model

    @get_model
    def drop_index(self, model: pw.Model, *columns: str) -> pw.Model:
        """Drop indexes."""
        columns_ = []
        for col in columns:
            field = model._meta.fields.get(col)
            if not field:
                continue

            if len(columns) == 1:
                field.unique = field.index = False

            if isinstance(field, pw.ForeignKeyField):
                col = col + "_id"
            columns_.append(col)
        index_name = make_index_name(model._meta.table_name, columns_)
        model._meta.indexes = [
            (cols, _) for (cols, _) in model._meta.indexes if columns != cols
        ]
        self.ops.append(self.migrator.drop_index(model._meta.table_name, index_name))
        return model

    @get_model
    def add_not_null(self, model: pw.Model, *names: str) -> pw.Model:
        """Add not null."""
        for name in names:
            field = model._meta.fields[name]
            field.null = False
            self.ops.append(
                self.migrator.add_not_null(model._meta.table_name, field.column_name)
            )
        return model

    @get_model
    def drop_not_null(self, model: pw.Model, *names: str) -> pw.Model:
        """Drop not null."""
        for name in names:
            field = model._meta.fields[name]
            field.null = True
            self.ops.append(
                self.migrator.drop_not_null(model._meta.table_name, field.column_name)
            )
        return model

    @get_model
    def add_default(self, model: pw.Model, name: str, default: t.Any) -> pw.Model:
        """Add default."""
        field = model._meta.fields[name]
        model._meta.defaults[field] = field.default = default
        self.ops.append(
            self.migrator.apply_default(model._meta.table_name, name, field)
        )
        return model


#  pylama:ignore=W0223,W0212,R



"""Automatically create migrations."""

import typing as t
from collections import OrderedDict
from collections.abc import Hashable

import peewee as pw
from playhouse.reflection import Column as VanilaColumn

INDENT = "    "
NEWLINE = "\n" + INDENT
FIELD_MODULES_MAP = {
    "ArrayField": "pw_pext",
    "BinaryJSONField": "pw_pext",
    "DateTimeTZField": "pw_pext",
    "HStoreField": "pw_pext",
    "IntervalField": "pw_pext",
    "JSONField": "pw_pext",
    "TSVectorField": "pw_pext",
}


def fk_to_params(field: pw.ForeignKeyField) -> t.Dict:
    """Get params from the given fk."""
    params = {}
    if field.on_delete is not None:
        params["on_delete"] = f"'{field.on_delete}'"

    if field.on_update is not None:
        params["on_update"] = f"'{field.on_update}'"

    return params


def dtf_to_params(field: pw.DateTimeField) -> t.Dict:
    """Get params from the given datetime field."""
    params = {}
    if not isinstance(field.formats, list):
        params["formats"] = field.formats

    return params


FIELD_TO_PARAMS = {
    pw.CharField: lambda f: {"max_length": f.max_length},
    pw.DecimalField: lambda f: {
        "max_digits": f.max_digits,
        "decimal_places": f.decimal_places,
        "auto_round": f.auto_round,
        "rounding": f.rounding,
    },
    pw.ForeignKeyField: fk_to_params,
    pw.DateTimeField: dtf_to_params,
}


class Column(VanilaColumn):
    """Get field's migration parameters."""

    def __init__(self, field: pw.Field, **kwargs):  # noqa
        super(Column, self).__init__(
            field.name,
            type(field),
            field.field_type,
            field.null,
            primary_key=field.primary_key,
            column_name=field.column_name,
            index=field.index,
            unique=field.unique,
            extra_parameters={},
        )
        if field.default is not None and not callable(field.default):
            self.default = repr(field.default)

        if self.field_class in FIELD_TO_PARAMS:
            self.extra_parameters.update(FIELD_TO_PARAMS[self.field_class](field))

        self.rel_model = None
        self.related_name = None
        self.to_field = None

        if isinstance(field, pw.ForeignKeyField):
            self.to_field = field.rel_field.name
            self.related_name = field.backref
            self.rel_model = (
                "'self'"
                if field.rel_model == field.model
                else "migrator.orm['%s']" % field.rel_model._meta.table_name
            )

    def get_field(self, space: str = " ") -> str:
        """Generate the field definition for this column."""
        field = super(Column, self).get_field()
        module = FIELD_MODULES_MAP.get(self.field_class.__name__, "pw")
        name, _, field = [s and s.strip() for s in field.partition("=")]
        return "{name}{space}={space}{module}.{field}".format(
            name=name, field=field, space=space, module=module
        )

    def get_field_parameters(self) -> t.Dict:
        """Generate parameters for self field."""
        params = super(Column, self).get_field_parameters()
        if self.default is not None:
            params["default"] = self.default
        return params


def diff_one(model1: pw.Model, model2: pw.Model, **kwargs) -> t.List[str]:
    """Find difference between given peewee models."""
    changes = []

    fields1 = model1._meta.fields
    fields2 = model2._meta.fields

    # Add fields
    names1 = set(fields1) - set(fields2)
    if names1:
        fields = [fields1[name] for name in names1]
        changes.append(create_fields(model1, *fields, **kwargs))

    # Drop fields
    names2 = set(fields2) - set(fields1)
    if names2:
        changes.append(drop_fields(model1, *names2))

    # Change fields
    fields_ = []
    nulls_ = []
    indexes_ = []
    for name in set(fields1) - names1 - names2:
        field1, field2 = fields1[name], fields2[name]
        diff = compare_fields(field1, field2)
        null = diff.pop("null", None)
        index = diff.pop("index", None)

        if diff:
            fields_.append(field1)

        if null is not None:
            nulls_.append((name, null))

        if index is not None:
            indexes_.append((name, index[0], index[1]))

    if fields_:
        changes.append(change_fields(model1, *fields_, **kwargs))

    for name, null in nulls_:
        changes.append(change_not_null(model1, name, null))

    for name, index, unique in indexes_:
        if index is True or unique is True:
            if fields2[name].unique or fields2[name].index:
                changes.append(drop_index(model1, name))
            changes.append(add_index(model1, name, unique))
        else:
            changes.append(drop_index(model1, name))

    # Check additional compound indexes
    indexes1 = model1._meta.indexes
    indexes2 = model2._meta.indexes

    # Drop compound indexes
    indexes_to_drop = set(indexes2) - set(indexes1)
    for index in indexes_to_drop:
        if isinstance(index[0], (list, tuple)) and len(index[0]) > 1:
            changes.append(drop_index(model1, name=index[0]))

    # Add compound indexes
    indexes_to_add = set(indexes1) - set(indexes2)
    for index in indexes_to_add:
        if isinstance(index[0], (list, tuple)) and len(index[0]) > 1:
            changes.append(add_index(model1, name=index[0], unique=index[1]))

    return changes


def diff_many(models1, models2, migrator=None, reverse=False):
    """Calculate changes for migrations from models2 to models1."""
    models1 = pw.sort_models(models1)
    models2 = pw.sort_models(models2)

    if reverse:
        models1 = reversed(models1)
        models2 = reversed(models2)

    models1 = OrderedDict([(m._meta.table_name, m) for m in models1])
    models2 = OrderedDict([(m._meta.table_name, m) for m in models2])

    changes = []

    for name, model1 in models1.items():
        if name not in models2:
            continue
        changes += diff_one(model1, models2[name], migrator=migrator)

    # Add models
    for name in [m for m in models1 if m not in models2]:
        changes.append(create_model(models1[name], migrator=migrator))

    # Remove models
    for name in [m for m in models2 if m not in models1]:
        changes.append(remove_model(models2[name]))

    return changes


def model_to_code(Model: pw.Model, **kwargs) -> str:
    """Generate migrations for the given model."""
    template = """class {classname}(pw.Model):
{fields}

{meta}
"""
    fields = INDENT + NEWLINE.join(
        [
            field_to_code(field, **kwargs)
            for field in Model._meta.sorted_fields
            if not (isinstance(field, pw.PrimaryKeyField) and field.name == "id")
        ]
    )
    meta = INDENT + NEWLINE.join(
        filter(
            None,
            [
                "class Meta:",
                INDENT + 'table_name = "%s"' % Model._meta.table_name,
                (INDENT + 'schema = "%s"' % Model._meta.schema)
                if Model._meta.schema
                else "",
                (
                    INDENT
                    + "primary_key = pw.CompositeKey{0}".format(
                        Model._meta.primary_key.field_names
                    )
                )
                if isinstance(Model._meta.primary_key, pw.CompositeKey)
                else "",
                (INDENT + "indexes = %s" % Model._meta.indexes)
                if Model._meta.indexes
                else "",
            ],
        )
    )

    return template.format(classname=Model.__name__, fields=fields, meta=meta)


def create_model(Model: pw.Model, **kwargs) -> str:
    """Generate migrations to create model."""
    return "@migrator.create_model\n" + model_to_code(Model, **kwargs)


def remove_model(Model: pw.Model, **kwargs) -> str:
    """Generate migrations to remove model."""
    return "migrator.remove_model('%s')" % Model._meta.table_name


def create_fields(Model: pw.Model, *fields: pw.Field, **kwargs) -> str:
    """Generate migrations to add fields."""
    return "migrator.add_fields(%s'%s', %s)" % (
        NEWLINE,
        Model._meta.table_name,
        NEWLINE
        + ("," + NEWLINE).join(
            [field_to_code(field, False, **kwargs) for field in fields]
        ),
    )


def drop_fields(Model: pw.Model, *fields: pw.Field, **kwargs) -> str:
    """Generate migrations to remove fields."""
    return "migrator.remove_fields('%s', %s)" % (
        Model._meta.table_name,
        ", ".join(map(repr, fields)),
    )


def field_to_code(field: pw.Field, space: bool = True, **kwargs) -> str:
    """Generate field description."""
    col = Column(field, **kwargs)
    return col.get_field(" " if space else "")


def compare_fields(field1: pw.Field, field2: pw.Field, **kwargs) -> t.Dict:
    """Find diffs between the given fields."""
    field_cls1, field_cls2 = type(field1), type(field2)
    if field_cls1 != field_cls2:  # noqa
        return {"cls": True}

    params1 = field_to_params(field1)
    params1["null"] = field1.null
    params2 = field_to_params(field2)
    params2["null"] = field2.null

    return dict(set(params1.items()) - set(params2.items()))


def field_to_params(field: pw.Field, **kwargs) -> t.Dict:
    """Generate params for the given field."""
    params = FIELD_TO_PARAMS.get(type(field), lambda f: {})(field)
    if (
        field.default is not None
        and not callable(field.default)
        and isinstance(field.default, Hashable)
    ):
        params["default"] = field.default

    params["index"] = field.index and not field.unique, field.unique

    params.pop("backref", None)  # Ignore backref
    return params


def change_fields(Model: pw.Model, *fields: pw.Field, **kwargs) -> str:
    """Generate migrations to change fields."""
    return "migrator.change_fields('%s', %s)" % (
        Model._meta.table_name,
        ("," + NEWLINE).join([field_to_code(f, False) for f in fields]),
    )


def change_not_null(Model: pw.Model, name: str, null: bool) -> str:
    """Generate migrations."""
    operation = "drop_not_null" if null else "add_not_null"
    return "migrator.%s('%s', %s)" % (operation, Model._meta.table_name, repr(name))


def add_index(
    Model: pw.Model, name: t.Union[str, t.Iterable[str]], unique: bool
) -> str:
    """Generate migrations."""
    return (
        f"migrator.add_index('{Model._meta.table_name}', {repr(name)}, unique={unique})"
    )


def drop_index(Model: pw.Model, name: t.Union[str, t.Iterable[str]]) -> str:
    """Generate migrations."""
    return f"migrator.drop_index('{Model._meta.table_name}', {repr(name)})"
